//! Abstract layer for Semtech SX1276/77/78/79.

use drone_cortex_m::peripherals::dma::Dma;
use drone_cortex_m::peripherals::spi::Spi;
use drone_cortex_m::peripherals::spi_dma::SpiDmaDx;
use drone_cortex_m::peripherals::timer::Timer;
use drone_cortex_m::reg::prelude::*;
use drone_cortex_m::thread::prelude::*;

/// Alias to refer `Conn::Spi`.
pub type ConnSpi<T> = <<T as Conn>::SpiDma as SpiDmaDx<
  <T as Conn>::ThreadLocal,
  <T as Conn>::IrqDmaTx,
  <T as Conn>::IrqDmaRx,
>>::Spi;

/// Alias to refer `Conn::DmaTx::Ccr`.
pub type ConnDmaTxCcr<T> =
  <<<T as Conn>::SpiDma as SpiDmaDx<
    <T as Conn>::ThreadLocal,
    <T as Conn>::IrqDmaTx,
    <T as Conn>::IrqDmaRx,
  >>::DmaTx as Dma<<T as Conn>::ThreadLocal, <T as Conn>::IrqDmaTx>>::Ccr;

/// Alias to refer `Conn::DmaRx::Ccr`.
pub type ConnDmaRxCcr<T> =
  <<<T as Conn>::SpiDma as SpiDmaDx<
    <T as Conn>::ThreadLocal,
    <T as Conn>::IrqDmaTx,
    <T as Conn>::IrqDmaRx,
  >>::DmaRx as Dma<<T as Conn>::ThreadLocal, <T as Conn>::IrqDmaRx>>::Ccr;

/// Generic Semtech SX1276/77/78/79.
pub trait Conn: Sized + Send + 'static {
  /// Concrete `Thread` implementation.
  type ThreadLocal: Thread;

  /// Transmitting DMA channel interrupt.
  type IrqDmaTx: ThreadNumber;

  /// Receiving DMA channel interrupt.
  type IrqDmaRx: ThreadNumber;

  /// Concrete `SpiDma` implementation.
  type SpiDma: SpiDmaDx<Self::ThreadLocal, Self::IrqDmaTx, Self::IrqDmaRx>;

  /// Initializes inner peripherals.
  fn init(&self);

  /// Lowers corresponding NSS pin.
  fn slave_select(&self);

  /// Rises corresponding NSS pin.
  fn slave_unselect(&self);

  /// `CR1` register configuration for SPI.
  fn spi_cr1<'a, 'b>(
    val: &'b mut <<ConnSpi<Self> as Spi>::Cr1 as RegRef<'a, Ftt>>::Hold,
  ) -> &'b mut <<ConnSpi<Self> as Spi>::Cr1 as RegRef<'a, Ftt>>::Hold;

  /// `CR2` register configuration for SPI.
  fn spi_cr2<'a, 'b>(
    val: &'b mut <<ConnSpi<Self> as Spi>::Cr2 as RegRef<'a, Ftt>>::Hold,
  ) -> &'b mut <<ConnSpi<Self> as Spi>::Cr2 as RegRef<'a, Ftt>>::Hold;

  /// `CCR` register configuration for transmitting DMA.
  fn dma_tx_ccr<'a, 'b>(
    val: &'b mut <ConnDmaTxCcr<Self> as RegRef<'a, Stt>>::Hold,
  ) -> &'b mut <ConnDmaTxCcr<Self> as RegRef<'a, Stt>>::Hold;

  /// `CCR` register configuration for receiving DMA.
  fn dma_rx_ccr<'a, 'b>(
    val: &'b mut <ConnDmaRxCcr<Self> as RegRef<'a, Stt>>::Hold,
  ) -> &'b mut <ConnDmaRxCcr<Self> as RegRef<'a, Stt>>::Hold;
}

/// Generic reset over timer `T`.
pub trait Reset<T: Timer + 'static>: Conn {
  /// Lowers NRESET pin.
  fn start_reset(&self);

  /// Rises NRESET pin.
  fn stop_reset(&self);

  /// Returns current value of the timer control register.
  fn timer_ctrl(timer: &T) -> <T::Ctrl as Reg<Ftt>>::Val;

  /// Returns a duration of reset procedure in timer ticks.
  fn reset_duration(&self, ctrl_val: <T::Ctrl as Reg<Ftt>>::Val)
    -> T::Duration;

  /// Returns a duration of post-reset wait in timer ticks.
  fn reset_wait_duration(
    &self,
    ctrl_val: <T::Ctrl as Reg<Ftt>>::Val,
  ) -> T::Duration;
}
