//! Semtech SX1276/77/78/79 generic tokens.
//!
//! See [`Tokens`] documentation for details.
//!
//! [`Tokens`]: struct.Tokens.html

use conn::{Conn, Reset};
use core::ops::{Deref, DerefMut};
use drone_cortex_m::peripherals::dma::Dma;
use drone_cortex_m::peripherals::spi::Spi;
use drone_cortex_m::peripherals::spi_dma::SpiDmaDx;
use drone_cortex_m::peripherals::timer::Timer;
use drone_cortex_m::reg::prelude::*;
use drone_sx1276::conn::{self, ResetError, SendError};

/// Semtech SX1276/77/78/79 generic tokens.
pub struct Tokens<C: Conn>(C);

/// SPI interface error.
#[derive(Debug, Clone, PartialEq, Eq)]
pub struct SpiDmaError;

/// Timer error.
#[derive(Debug, Clone, PartialEq, Eq)]
pub struct TimerError;

impl<C: Conn> Tokens<C> {
  /// Wraps the application-specific `tokens`.
  #[inline(always)]
  pub fn new(tokens: C) -> Self {
    Tokens(tokens)
  }

  /// Converts to inner application-specific tokens.
  #[inline(always)]
  pub fn into_tokens(self) -> C {
    self.0
  }
}

impl<C: Conn> conn::Conn for Tokens<C> {
  type Spi = C::SpiDma;
  type SendErrorKind = SpiDmaError;

  #[inline(always)]
  fn init(&self) {
    self.0.init();
  }

  #[inline(always)]
  unsafe fn transfer(
    self,
    spi: Self::Spi,
    address: u8,
    count: usize,
    tx_buf: *const u8,
    rx_buf: *mut u8,
  ) -> Box<Future<Item = (Self, Self::Spi), Error = SendError<Self>> + Send> {
    spi.dma_rx().set_memory_address(rx_buf as usize);
    spi.dma_rx().set_number_of_data(count as u16 + 1);
    spi.dma_rx().ccr().reset(C::dma_rx_ccr);
    spi.dma_tx().set_memory_address(tx_buf as usize);
    spi.dma_tx().set_number_of_data(count as u16);
    spi.dma_tx().ccr().reset(C::dma_tx_ccr);
    self.0.slave_select();
    spi.spi().send_byte(address);
    let cr1 = C::spi_cr1(&mut spi.spi().cr1().default()).val();
    let cr2 = C::spi_cr2(&mut spi.spi().cr2().default()).val();
    let spi = spi.transfer_complete(cr1, cr2);
    Box::new(AsyncFuture::new(move || match await!(spi) {
      Ok(spi) => {
        spi.dma_rx().ccr().store_default();
        spi.dma_tx().ccr().store_default();
        spi.spi().busy_wait();
        spi.spi().cr1().store_default();
        spi.spi().cr2().store_default();
        self.0.slave_unselect();
        Ok((self, spi))
      }
      Err(spi) => Err(SendError {
        tokens: self,
        spi,
        kind: SpiDmaError,
      }),
    }))
  }
}

impl<C: Reset<T>, T: Timer + 'static> conn::Reset<T> for Tokens<C> {
  type ResetErrorKind = TimerError;

  #[inline(always)]
  fn reset(
    self,
    mut timer: T,
  ) -> Box<Future<Item = (Self, T), Error = ResetError<Self, T>> + Send> {
    macro_rules! try {
      ($e:expr) => {
        if let Err(()) = $e {
          return Err(ResetError {
            tokens: self,
            timer,
            kind: TimerError,
          });
        }
      }
    }
    Box::new(AsyncFuture::new(move || {
      self.0.start_reset();
      try!(await!({
        let ctrl = C::timer_ctrl(&timer);
        timer.sleep(self.0.reset_duration(ctrl), ctrl)
      }));
      self.0.stop_reset();
      try!(await!({
        let ctrl = C::timer_ctrl(&timer);
        timer.sleep(self.0.reset_wait_duration(ctrl), ctrl)
      }));
      Ok((self, timer))
    }))
  }
}

impl<C: Conn> Deref for Tokens<C> {
  type Target = C;

  #[inline(always)]
  fn deref(&self) -> &C {
    &self.0
  }
}

impl<C: Conn> DerefMut for Tokens<C> {
  #[inline(always)]
  fn deref_mut(&mut self) -> &mut C {
    &mut self.0
  }
}
